using System;
using System.Text;
using System.Security.Cryptography;
using WingServerCommon.Config;
using System.Collections.Generic;
using System.Linq;
using System.Globalization;
using WingServerCommon.Config.Parameters;

namespace WingTokenModule.Common
{
    /// <summary>
    /// 签名算法
    /// </summary>
    public class SignatureHelper
    {
        /// <summary>
        /// Vod 云点播 签名生成
        /// </summary>
        /// <returns>签名字符串</returns>
        public static string GetVodAuthorization()
        {
            var m_strSecId = EnvironmentConfigManager.GetParammeter<StringParameter>("VodPlayer", "AppKey").Value;
            var m_strSecKey = EnvironmentConfigManager.GetParammeter<StringParameter>("VodPlayer", "AppSecret").Value;
            // string m_strSecId = appKey;
            // string m_strSecKey = appSecret;
            var strContent = "";
            var m_qwNowTime = GetIntTimeStamp();
            var m_iSignValidDuration = 3600 * 24 * 2;// 签名有效期：2天
            var m_iRandom = new Random().Next(0, 1000000);//随机数
            strContent += ("secretId=" + Uri.EscapeDataString((m_strSecId)));
            strContent += ("&currentTimeStamp=" + m_qwNowTime);
            strContent += ("&expireTime=" + (m_qwNowTime + m_iSignValidDuration));
            strContent += ("&random=" + m_iRandom);
            byte[] bytesSign = GetByteHashHmac(strContent, m_strSecKey);
            byte[] byteContent = System.Text.Encoding.Default.GetBytes(strContent);
            byte[] nCon = new byte[bytesSign.Length + byteContent.Length];
            bytesSign.CopyTo(nCon, 0);
            byteContent.CopyTo(nCon, bytesSign.Length);
            return Convert.ToBase64String(nCon);
        }

        /// <summary>
        /// 权鉴 签名地址生成
        /// </summary>
        /// <param name="fileType">上传文件类型</param>
        /// <returns>签名字符串</returns>
        public static string GetStorageAuthorization(string fileName, WingInterfaceLibrary.Enum.StorageServerEnum serverType, string method = "put",
            List<KeyValuePair<string, string>> _requestParams = null, List<KeyValuePair<string, string>> _headers = null, bool isCompleted = false)
        {
            var storageSettingsParam = ConfigurationManager.StorageSettings;
            if (storageSettingsParam == null)
            {
                return string.Empty;
            }
            var now = DateTimeOffset.Now;
            // long startTimeStamp = now.ToUnixTimeSeconds();
            // long endTimeStamp = now.AddDays(1).ToUnixTimeSeconds();
            long startTimeStamp = 0;
            long endTimeStamp = 0;
            if (isCompleted)
            {
                startTimeStamp = now.AddDays(-1).ToUnixTimeSeconds();
                endTimeStamp = now.AddDays(2).ToUnixTimeSeconds();
            }
            else
            {
                startTimeStamp = now.ToUnixTimeSeconds();
                endTimeStamp = now.AddDays(1).ToUnixTimeSeconds();
            }
            var keyTime = $"{startTimeStamp};{endTimeStamp}";
            var signKey = HMACSHA1(storageSettingsParam.Sercret, keyTime);

            var urlParamListBuilder = new StringBuilder();
            var stringParamBuilder = new StringBuilder();
            var urlParamList = string.Empty;
            var httpParameters = string.Empty;
            if (_requestParams?.Count > 0)
            {
                _requestParams = _requestParams.OrderBy(c => c.Key).ToList();//TODO check		
                foreach (var param in _requestParams)
                {
                    var encodeKey = Encode(param.Key).ToLower();
                    urlParamListBuilder.Append($"{encodeKey};");
                    var encodeValue = Encode(param.Value);
                    stringParamBuilder.Append($"{encodeKey}={encodeValue}&");
                }

                urlParamList = urlParamListBuilder.ToString();
                if (!string.IsNullOrEmpty(urlParamList))
                {
                    urlParamList = urlParamList.TrimEnd(';');
                }
                httpParameters = stringParamBuilder.ToString();
                if (!string.IsNullOrEmpty(httpParameters))
                {
                    httpParameters = httpParameters.TrimEnd('&');
                }
            }

            var httpHeaders = string.Empty;
            var headerListBuilder = new StringBuilder();
            var httpHeadersBuilder = new StringBuilder();
            var headerList = string.Empty;
            if (_headers?.Count > 0)
            {
                _headers = _headers.OrderBy(c => c.Key).ToList();//TODO check
                foreach (var param in _headers)
                {
                    var encodeKey = Encode(param.Key).ToLower();
                    headerListBuilder.Append($"{encodeKey};");
                    var encodeValue = Encode(param.Value);
                    httpHeadersBuilder.Append($"{encodeKey}={encodeValue}&");
                }
                headerList = headerListBuilder.ToString();
                if (!string.IsNullOrEmpty(headerList))
                {
                    headerList = headerList.TrimEnd(';');
                }
                httpHeaders = httpHeadersBuilder.ToString();
                if (!string.IsNullOrEmpty(httpHeaders))
                {
                    httpHeaders = httpHeaders.TrimEnd('&');
                }
            }

            //Generatet HttpString
            //string httpString = string.Format($"put\n/{fileName}\n\n\n");
            string httpString = string.Format($"{method}\n/{fileName}\n{httpParameters}\n{httpHeaders}\n");
            //Generate StringToSign
            string stringToSign = string.Format("sha1\n{0}\n{1}\n", keyTime, SHA1Format(httpString));
            //Generate Signature
            var signature = "";
            if (serverType == WingInterfaceLibrary.Enum.StorageServerEnum.VCS)
            {
                var newStringToSign = MD5.GetHashString(stringToSign, Encoding.UTF8).ToLower();
                signature = HMACSHA1(signKey, newStringToSign);
            }
            else
            {
                signature = HMACSHA1(signKey, stringToSign);
            }
            var authorization = "";
            if (!string.IsNullOrEmpty(headerList) && !string.IsNullOrEmpty(urlParamList))
            {
                authorization = $"q-sign-algorithm=sha1&q-ak={storageSettingsParam.AppId}&q-sign-time={keyTime}&q-key-time={keyTime}&q-header-list={headerList}&q-url-param-list={urlParamList}&q-signature={signature}";
            }
            else if (!string.IsNullOrEmpty(headerList) && string.IsNullOrEmpty(urlParamList))
            {
                authorization = $"q-sign-algorithm=sha1&q-ak={storageSettingsParam.AppId}&q-sign-time={keyTime}&q-key-time={keyTime}&q-header-list={headerList}&q-url-param-list=&q-signature={signature}";
            }
            else if (string.IsNullOrEmpty(headerList) && !string.IsNullOrEmpty(urlParamList))
            {
                authorization = $"q-sign-algorithm=sha1&q-ak={storageSettingsParam.AppId}&q-sign-time={keyTime}&q-key-time={keyTime}&q-header-list=&q-url-param-list={urlParamList}&q-signature={signature}";
            }
            else
            {
                authorization = $"q-sign-algorithm=sha1&q-ak={storageSettingsParam.AppId}&q-sign-time={keyTime}&q-key-time={keyTime}&q-header-list=&q-url-param-list=&q-signature={signature}";
            }
            return authorization;
        }

        /// <summary>
        /// 加密算法
        /// </summary>
        /// <param name="content">请求参数<param>
        /// <returns>加密结果</returns>
        private static string SHA1Format(string content)
        {
            var sha1 = SHA1.Create();
            byte[] c = Encoding.Default.GetBytes(content);
            byte[] sc = sha1.ComputeHash(c);
            StringBuilder sb = new StringBuilder("");
            foreach (byte b in sc)
            {
                sb.AppendFormat("{0:x2}", b);
            }
            return sb.ToString();
        }

        /// <summary>
        /// 加密算法
        /// </summary>
        /// <param name="secret">秘钥</param>
        /// <param name="content">时间戳<param>
        /// <returns>上传生成结果对象</returns>
        private static string HMACSHA1(string secret, string content)
        {
            byte[] keyByte = Encoding.Default.GetBytes(secret);
            HMACSHA1 hmacsha1 = new HMACSHA1(keyByte);
            byte[] messageBytes = Encoding.Default.GetBytes(content);
            byte[] hashmessage = hmacsha1.ComputeHash(messageBytes);
            StringBuilder sb = new StringBuilder("");
            foreach (byte b in hashmessage)
            {
                sb.AppendFormat("{0:x2}", b);
            }
            return sb.ToString();
        }

        /// <summary>
        /// 编码
        /// </summary>
        /// <param name="value">值</param>
        /// <returns>编码结果</returns>
        public static string Encode(string value)
        {
            return Encode(value, Encoding.UTF8);
        }

        /// <summary>
        /// 针对单个value，则只需满足 URLAllowChars 不需要编码即可.
        /// urlEncode: 转为一个byte -> 转为两个16进制 -> 前面加上 %
        /// </summary>
        /// <param name="value"></param>
        /// <param name="encoding"></param>
        /// <returns></returns>
        public static string Encode(string value, Encoding encoding)
        {
            //只有字母和数字[0-9a-zA-Z]、一些特殊符号"-_.!*~',以及某些保留字，才可以不经过编码直接用于URL。"
            string URLAllowChars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789-_.~";
            if (String.IsNullOrEmpty(value))
            {

                return String.Empty;
            }
            // %xy%xy
            // %xy%xy
            StringBuilder result = new StringBuilder(value.Length * 2);
            byte[] strToBytes = encoding.GetBytes(value);

            foreach (byte b in strToBytes)
            {
                char ch = (char)b;

                if (URLAllowChars.IndexOf(ch) != -1)
                {
                    result.Append(ch);
                }
                else
                {
                    result.Append('%').Append(String.Format(CultureInfo.InvariantCulture, "{0:X2}", (int)b));
                }
            }

            return result.ToString();
        }

        /// <summary>
        /// 时间秒数时间戳
        /// </summary>
        /// <returns></returns>
        private static long GetIntTimeStamp()
        {
            TimeSpan ts = DateTime.UtcNow - new DateTime(1970, 1, 1);
            return Convert.ToInt64(ts.TotalSeconds);
        }

        /// <summary>
        /// 加密算法
        /// </summary>
        /// <param name="signatureString"></param>
        /// <param name="secretKey"></param>
        /// <returns></returns>
        private static byte[] GetByteHashHmac(string signatureString, string secretKey)
        {
            var enc = Encoding.UTF8;
            HMACSHA1 hmac = new HMACSHA1(enc.GetBytes(secretKey));
            hmac.Initialize();
            byte[] buffer = enc.GetBytes(signatureString);
            return hmac.ComputeHash(buffer);
        }

    }
}